/*
 * Copyright (c) 2014, 2015, Oracle and/or its affiliates. All rights reserved.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation.
 *
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Please contact Oracle, 500 Oracle Parkway, Redwood Shores, CA 94065 USA
 * or visit www.oracle.com if you need additional information or have any
 * questions.
 */
package com.oracle.truffle.r.runtime.conn;

import java.io.*;
import java.nio.*;
import java.util.*;

import com.oracle.truffle.api.source.*;
import com.oracle.truffle.r.runtime.*;
import com.oracle.truffle.r.runtime.conn.ConnectionSupport.BaseRConnection;
import com.oracle.truffle.r.runtime.conn.ConnectionSupport.*;
import com.oracle.truffle.r.runtime.data.*;
import com.oracle.truffle.r.runtime.data.model.*;
import com.oracle.truffle.r.runtime.env.*;
import com.oracle.truffle.r.runtime.env.REnvironment.PutException;

public class TextConnections {
    public static class TextRConnection extends BaseRConnection {
        protected String nm;
        protected RAbstractStringVector object;
        protected REnvironment env;

        public TextRConnection(String nm, RAbstractStringVector object, REnvironment env, String modeString) throws IOException {
            super(ConnectionClass.Text, modeString, AbstractOpenMode.Read);
            this.nm = nm;
            this.object = object;
            this.env = env;
            openNonLazyConnection();
        }

        @Override
        public String getSummaryDescription() {
            return nm;
        }

        @Override
        protected void createDelegateConnection() throws IOException {
            DelegateRConnection delegate = null;
            switch (getOpenMode().abstractOpenMode) {
                case Read:
                    delegate = new TextReadRConnection(this);
                    break;
                case Write:
                    delegate = new TextWriteRConnection(this);
                    break;
                default:
                    throw RError.nyi((SourceSection) null, "unimplemented open mode: " + getOpenMode().modeString);
            }
            setDelegate(delegate);
        }

        public String[] getValue() {
            return ((GetConnectionValue) theConnection).getValue();
        }
    }

    private interface GetConnectionValue {
        String[] getValue();
    }

    private static class TextReadRConnection extends DelegateReadRConnection implements GetConnectionValue {
        private String[] lines;
        private int index;

        TextReadRConnection(TextRConnection base) {
            super(base);
            StringBuffer sb = new StringBuffer();
            for (int i = 0; i < base.object.getLength(); i++) {
                sb.append(base.object.getDataAt(i));
                // vector elements are implicitly terminated with a newline
                sb.append('\n');
            }
            lines = sb.toString().split("\\n");
        }

        @Override
        public String[] readLinesInternal(int n) throws IOException {
            int nleft = lines.length - index;
            int nlines = nleft;
            if (n > 0) {
                nlines = n > nleft ? nleft : n;
            }

            String[] result = new String[nlines];
            System.arraycopy(lines, index, result, 0, nlines);
            index += nlines;
            return result;
        }

        @SuppressWarnings("hiding")
        @Override
        public void writeLines(RAbstractStringVector lines, String sep) throws IOException {
            throw new IOException(RError.Message.CANNOT_WRITE_CONNECTION.message);
        }

        @Override
        public InputStream getInputStream() throws IOException {
            throw RInternalError.shouldNotReachHere();
        }

        @Override
        public void closeAndDestroy() throws IOException {
            base.closed = true;
        }

        @Override
        public void close() {
        }

        @Override
        public int readBin(ByteBuffer buffer) throws IOException {
            throw RError.nyi(null, " readBin on text connection");
        }

        @Override
        public byte[] readBinChars() throws IOException {
            throw RError.nyi(null, " readBinChars on text connection");
        }

        @Override
        public String readChar(int nchars, boolean useBytes) throws IOException {
            throw RError.nyi(null, " readChar on text connection");
        }

        public String[] getValue() {
            throw RError.nyi(null, " textConnectionValue");
        }

    }

    private static class TextWriteRConnection extends DelegateWriteRConnection implements GetConnectionValue {
        String incompleteLine;
        final RStringVector textVec;

        protected TextWriteRConnection(BaseRConnection base) {
            super(base);
            TextRConnection textBase = (TextRConnection) base;
            try {
                textVec = RDataFactory.createStringVector(0);
                textBase.env.put(textBase.nm, textVec);
            } catch (PutException ex) {
                throw RError.error((SourceSection) null, ex);
            }
            // lock the binding until close
            textBase.env.lockBinding(textBase.nm);
        }

        @Override
        public OutputStream getOutputStream() throws IOException {
            throw RInternalError.shouldNotReachHere();
        }

        @Override
        public void closeAndDestroy() throws IOException {
            base.closed = true;
            TextRConnection textBase = (TextRConnection) base;
            textBase.env.unlockBinding(textBase.nm);
        }

        @Override
        public void close() throws IOException {
        }

        @Override
        public void writeLines(RAbstractStringVector lines, String sep) throws IOException {
            StringBuffer sb = new StringBuffer();
            if (incompleteLine != null) {
                sb.append(incompleteLine);
                incompleteLine = null;
            }
            for (int i = 0; i < lines.getLength(); i++) {
                sb.append(lines.getDataAt(i));
                sb.append(sep);
            }
            String result = sb.toString();
            int nlIndex;
            int px = 0;
            ArrayList<String> appendedLines = new ArrayList<>();
            while ((nlIndex = result.indexOf('\n', px)) >= 0) {
                appendedLines.add(result.substring(px, nlIndex));
                nlIndex++;
                px = nlIndex;
            }
            if (px < result.length()) {
                incompleteLine = result.substring(px);
            }
            if (appendedLines.size() > 0) {
                // update the vector data
                String[] appendedData = new String[appendedLines.size()];
                appendedLines.toArray(appendedData);
                String[] existingData = textVec.getDataWithoutCopying();
                String[] updateData = appendedData;
                if (existingData.length > 0) {
                    updateData = new String[existingData.length + appendedData.length];
                    System.arraycopy(existingData, 0, updateData, 0, existingData.length);
                    System.arraycopy(appendedData, 0, updateData, existingData.length, appendedData.length);
                }
                textVec.setDataInternal(updateData);
            }
        }

        @Override
        public void flush() throws IOException {
        }

        @Override
        public void writeString(String s, boolean nl) throws IOException {
            // TODO Auto-generated method stub

        }

        @Override
        public void writeChar(String s, int pad, String eos, boolean useBytes) throws IOException {
            throw RError.nyi(null, " writeChar on text connection");
        }

        @Override
        public void writeBin(ByteBuffer buffer) throws IOException {
            throw RError.nyi(null, " writeBin on text connection");
        }

        public String[] getValue() {
            throw RError.nyi(null, " textConnectionValue");
        }

    }

}
